#!/usr/bin/perl -w

# ann_feats2.pl gets an annotation file from fasta36 -V with a line of the form:

# gi|62822551|sp|P00502|GSTA1_RAT Glutathione S-transfer\n  (at least from pir1.lseg)
#
# it must:
# (1) read in the line
# (2) parse it to get the up_acc
# (3) return the tab delimited features
#

# this version can read feature2 uniprot features (acc/pos/end/label/value), but returns sorted start/end domains

use strict;

use DBI;
use Getopt::Long;
use Pod::Usage;

use vars qw($host $db $port $user $pass);

my $hostname = `hostname`;

unless ($hostname =~ m/ebi/) {
#  ($host, $db, $port, $user, $pass)  = ("wrpxdb.bioch.virginia.edu", "uniprot", 3306, "web_user", "fasta_www");
  ($host, $db, $port, $user, $pass)  = ("xdb", "uniprot", 3306, "web_user", "fasta_www");
}
else {
  ($host, $db, $port, $user, $pass)  = ("mysql-pearson", "up_db", 4124, "web_user", "fasta_www");
}

my ($sstr, $lav, $no_doms, $no_feats, $shelp, $help) = (0,0,0,0,0,0);

GetOptions(
    "host=s" => \$host,
    "db=s" => \$db,
    "user=s" => \$user,
    "password=s" => \$pass,
    "port=i" => \$port,
    "lav" => \$lav,
    "no_doms" => \$no_doms,
    "no-doms" => \$no_doms,
    "nodoms" => \$no_doms,
    "no_feats" => \$no_feats,
    "no-feats" => \$no_feats,
    "nofeats" => \$no_feats,
    "sstr" => \$sstr,
    "h|?" => \$shelp,
    "help" => \$help,
    );

pod2usage(1) if $shelp;
pod2usage(exitstatus => 0, verbose => 2) if $help;
pod2usage(1) unless @ARGV;

my $connect = "dbi:mysql(AutoCommit=>1,RaiseError=>1):database=$db";
$connect .= ";host=$host" if $host;
$connect .= ";port=$port" if $port;

my $dbh = DBI->connect($connect,
		       $user,
		       $pass
		      ) or die $DBI::errstr;

my @feat_keys = qw(ACT_SITE MOD_RES BINDING SITE METAL VARIANT MUTAGEN);
my @feat_vals = ( '=','*','#','^','!','V','V');

my @dom_keys = qw( DOMAIN REPEAT );
my @dom_vals = ( [ '[', ']'],[ '[', ']']);

my @ssr_keys = qw( SSTR );
my @ssr_vals = ( [ '[', ']']);

my %annot_types = ();

my $get_annot_sub = \&get_fasta_annots;
if ($lav) {
  $no_feats = 1;
  $get_annot_sub = \&get_lav_annots;
}

if ($sstr) {@annot_types{@ssr_keys} = @ssr_vals;}
else {
  @annot_types{@feat_keys} = @feat_vals unless ($no_feats);
  @annot_types{@dom_keys} = @dom_vals unless ($no_doms);
}

my $get_annots_id = $dbh->prepare('select * from features2 join annot on(up_acc=acc) where name=? order by pos');
my $get_annots_acc = $dbh->prepare('select * from features2 where acc=? order by pos');
my $get_annots = $get_annots_id;

my ($tmp, $gi, $sdb, $acc, $id, $use_acc);

unless ( $no_feats || $sstr) {
  print "=*:phosphorylation\n";
  print "==".":active site\n";
  print "=@".":site\n";
  print "=^:binding\n";
  print "=!:metal binding\n";
}

# get the query
my $query = shift @ARGV;

$query =~ s/^>//;

my $ANN_F;

#if it's a file I can open, read and parse it
if ($query !~ m/\|/ && open($ANN_F, $query)) {

  while (my $a_line = <$ANN_F>) {
    $a_line =~ s/^>//;
    chomp $a_line;
    show_annots($a_line, $get_annot_sub);
  }
}
else {
  show_annots($query, $get_annot_sub);
}
exit(0);

sub show_annots {
  my ($annot_line, $get_annot_sub) = @_;


  print ">$annot_line\n";

  if ($annot_line =~ m/^gi\|/) {
    $use_acc = 1;
    ($tmp, $gi, $sdb, $acc, $id) = split(/\|/,$annot_line);
  }
  elsif ($annot_line =~ m/SP:(\w+)/) {
    $use_acc = 0;
    $sdb = 'sp';;
    $id = $1;
  }
  elsif ($annot_line =~ m/TR:(\w+)/) {
    $use_acc = 0;
    $sdb = 'tr';
    $id = $1;
  }
  else {
    $use_acc = 1;
    ($sdb, $acc, $id) = split(/\|/,$annot_line);
  }


  # remove version number
  unless ($use_acc) {
    $get_annots = $get_annots_id;
    $get_annots->execute($id);
  }
  else {
    $get_annots = $get_annots_acc;
    $acc =~ s/\.\d+$//;
    $get_annots->execute($acc);
  }

  $get_annot_sub->(\%annot_types, $get_annots);
}

sub get_fasta_annots {
  my ($annot_types, $get_annots) = @_;

  my ($pos, $end, $label, $value, $comment);

  my @feats = ();

  while (($acc, $pos, $end, $label, $value) = $get_annots->fetchrow_array()) {
    if ($annot_types->{$label}) {
      if ($label =~ m/VARIANT/) {
	my ($aa_res, $comment) = split(/\(/,$value);
	if ($comment) {	
	    $comment =~ s/\)//;
# remove the  /FTId=VAR_014497 information
	    $comment =~ s/\s+\/FTId=.*$//;
	}
	else {$comment = "";}
	next if ($comment =~ /MISSING/);
	my ($vfrom, $vto) = ($aa_res =~ m/(\w)\s*->\s*(\w)/);
	if ($vto) {
	  $comment = '' unless $comment;
	  $value = $vto;
	  push @feats, [$pos, $annot_types->{$label}, $value, $comment];
	}
      } elsif ($label =~ m/MUTAGEN/) {
	my ($aa_res, $comment) = split(/: /,$value);
	next if ($comment =~ /MISSING/);
	my ($vfrom, $vto) = split(/\->/,$aa_res);
	if ($vto) {
	  my @vto_list = split(/,/,$vto);
	  $value = $vto;
	  for my $val ( @vto_list) {
	    push @feats, [$pos, $annot_types->{$label}, $val, "Mutagen: $comment"];
	  }
	}
      } elsif ($label =~ m/DOMAIN/ || $label =~ m/REPEAT/) {
	push @feats, [$pos, $annot_types->{$label}[0], "-", "$value"];
	push @feats, [$end, $annot_types->{$label}[1], "-", ""];

      } elsif ($label =~ m/SSTR/) {
	next if $value =~ m/TURN/i;
	push @feats, [$pos, $annot_types->{$label}[0], "-", "$value"];
	push @feats, [$end, $annot_types->{$label}[1], "-", ""];
      }
      else {
#	print join("\t",($pos, $annot_types->{$label})),"\n";
#	print join("\t",($pos, $annot_types->{$label}, "-", "$label: $value")),"\n";
	push @feats, [$pos, $annot_types->{$label}, "-", "$label: $value"];
      }
    }
  }

  @feats = sort { $a->[0] <=> $b->[0] } @feats;
  for my $feat ( @feats ) {
    print join("\t",@$feat)."\n";
  }
}

sub get_lav_annots {
  my ($annot_types, $get_annots) = @_;

  my ($pos, $end, $label, $value, $comment);

  my %annot = ();
  while (($acc, $pos, $end, $label, $value) = $get_annots->fetchrow_array()) {
    next unless ($annot_types->{$label});
#    next unless ($label =~ m/^DOMAIN/ || $label =~ m/^REPEAT/);
    %annot = ();
    $annot{beg} = $pos;
    $annot{descr} = $value;
    $annot{end} = $end;
    print join("\t",($annot{beg}, $annot{end}, $annot{descr})),"\n";
  }
}

__END__

=pod

=head1 NAME

ann_feats2.pl

=head1 SYNOPSIS

 ann_feats2.pl --no_doms --no_feats --lav 'sp|P09488|GSTM1_NUMAN' | accession.file

=head1 OPTIONS

 -h	short help
 --help include description
 --no-doms  do not show domain boundaries (domains are always shown with --lav)
 --no-feats do not show feature (variants, active sites, phospho-sites)
 --lav  produce lav2plt.pl annotation format, only show domains/repeats

 --host, --user, --password, --port --db -- info for mysql database

=head1 DESCRIPTION

C<ann_feats2.pl> extracts feature, domain, and repeat information from
a msyql database (default name, uniprot) built by parsing the
uniprot_sprot.dat and uniprot_trembl.dat feature tables.  Given a
command line argument that contains a sequence accession (P09488) or
identifier (GSTM1_HUMAN), the program looks up the features available
for that sequence and returns them in a tab-delimited format:

 >sp|P09488
 2	-	88	DOMAIN: GST N-terminal.
 7	V	F	Mutagen: Reduces catalytic activity 100- fold.
 23	*	-	MOD_RES: Phosphotyrosine (By similarity).
 33	*	-	MOD_RES: Phosphotyrosine (By similarity).
 34	*	-	MOD_RES: Phosphothreonine (By similarity).
 90	-	208	DOMAIN: GST C-terminal.
 108	V	S	Mutagen: Changes the properties of the enzyme toward some substrates.
 108	V	Q	Mutagen: Reduces catalytic activity by half.
 109	V	I	Mutagen: Reduces catalytic activity by half.
 116	#	-	BINDING: Substrate.
 116	V	A	Mutagen: Reduces catalytic activity 10-fold.
 116	V	F	Mutagen: Slight increase of catalytic activity.
 173	V	N	in allele GSTM1B; dbSNP:rs1065411.
 210	V	T	in dbSNP:rs449856.

If features are provided, then a legend of feature symbols is provided
as well:

 =*:phosphorylation
 ==:active site
 =@:site
 =^:binding
 =!:metal binding

If the C<--lav> option is specified, domain and repeat features are
presented in a different format for the C<lav2plt.pl> program:

  >sp|P09488|GSTM1_HUMAN
  2	88	GST N-terminal.
  90	208	GST C-terminal.

C<ann_feats2.pl> is designed to be used by the B<FASTA> programs with
the C<-V \!ann_feats2.pl> option.  It can also be used with the lav2plt.pl
program with the C<--xA "\!ann_feats2.pl --lav"> or C<--yA "\!ann_feats2.pl --lav"> options.

=head1 AUTHOR

William R. Pearson, wrp@virginia.edu

=cut
